"""QAM is a simple RPC Framework which uses the AMQP Protocoll as Transport Protocoll."""
VERSION = (0, 2, 0)
__version__ = ".".join(map(str, VERSION))
__author__ = "QAM Team"
__contact__ = "qamteam@gmail.com"
__homepage__ = "http://packages.python.org/qam"
__docformat__ = "restructuredtext"

